/*
 *  Smart Cache , http proxy caching server
 *  Copyright (C) 1998-9 Radim Kolar <hsn@cybermail.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA or
 *  download copy from http://www.gnu.org/copyleft/gpl.html
 *
 */

import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.net.*;

public final class cachedir{

private String localdir;

private Hashtable objects;
private boolean dirty;

public static final String DIRINFO=".cacheinfo";
static mgr parent;

cachedir(String locdir)
{
 localdir=locdir;
 File f=new File(locdir);
 if(f.exists() && !f.isDirectory())
  {
   /* konflikt se jmenem souboru, smazeme ho */ 
   System.out.println("Directory "+locdir+" has the same name as existing file. File removed.");
   f.delete();
  }
 if(!f.exists()) 
                  {
                   f.mkdirs();
                   if(!f.exists()) 
                     { 
                       /* likvidujeme up-dirs, mozna je to konflikt se souborem */
                       File thisdir=new File(localdir.substring(0,localdir.length()-1));
                       while(true)
                       {  
                          String par;
                          thisdir.delete();
                          par=thisdir.getParent();
                          if(par==null) break;
                          thisdir=new File(par);
                       }
                       f.mkdirs();
                       if(!f.exists())
                       {     
                       System.out.println("Cannot create directory "+localdir+" - turning caching off for it.");
                       localdir=null;
                       objects=new Hashtable();
                       return;
                       // throw new NullPointerException("Cannot create directory "+localdir+" - This is parser bug. REPORT TO AUTHOR!");/
                       }
                     }
                   
                  }
                  
 objects=new Hashtable();
 dirty=false;
 /* nacteme cachovane objekty */
 
 DataInputStream is=null;
 try{
 is=new DataInputStream( 
                 new BufferedInputStream(
                 new FileInputStream( localdir+DIRINFO ),4096));
                 
 int howmany=is.readInt();
 for(int i=0;i<howmany;i++)
  {
   cacheobject o;
   o=new cacheobject(is,this);
   objects.put(o.getName(),o);
  }
 is.close(); 
 }
 catch(IOException e) {
                       try{
                         if(is!=null) { System.out.println("I/O error reading "+localdir+DIRINFO+" ,file corrupted?");
                                        is.close();
                                      }
                          }
                        catch (IOException z) {}  
                       };
} /* konstruktor */

final public void setDirty()
{
 dirty=true;
}

final public boolean isDirty()
{
 return dirty;
}
final public synchronized void save()
{
 if(localdir==null) { dirty=false;return;}
 if(!dirty) return;
 
 /* smazat ty, ktere neukladame */
 Enumeration en=objects.elements();
 while(en.hasMoreElements())
  {
   cacheobject o=(cacheobject)en.nextElement();
   if(!o.needSave())
                      objects.remove(o.getName());

  }
 int howmany;
 howmany=objects.size();
 if(howmany==0) {
                  /* nemusime nic ukladat */ 
                  dirty=false;
                  return;
                 }
 
  try{
 DataOutputStream os=new DataOutputStream(
                     new BufferedOutputStream(
                     new FileOutputStream( localdir+DIRINFO),4096 ));
                     
 os.writeInt(howmany);
 en=objects.elements();
 while(en.hasMoreElements())
  {
   cacheobject o=(cacheobject)en.nextElement();
   if(o.needSave())
                o.save(os);
       else
                 { objects.remove(o.getName());
                   os.close();save();return;}
  }
 os.close();
 dirty=false;
 }
 
 catch (FileNotFoundException e) 
    /* nekdo nam smazal hot dir, ze by BFU nebo GC ? */
   {
     System.out.println("Lost Directory "+localdir);
     localdir=null;
     objects=new Hashtable();
     dirty=false;
     return;
    }
 catch (IOException e) {System.out.println("Directory "+localdir+" save error "+e);}
}

final public boolean cleandir()
{
  if(localdir==null) return false;
  if(objects.size()!=0) return false; /* safe-check */
  boolean rc;
  rc=true;
  
   File thisdir=new File(localdir.substring(0,localdir.length()-1));
   String filez[]=thisdir.list();
   if(filez==null) return true;
   for(int i=0;i<filez.length;i++)
   {
    if(!new File(localdir,filez[i]).delete()) rc=false;
   }
   if(rc) System.err.println("Cleaned empty dir: "+localdir);
     else
       { localdir=null;return false;}
   
   /* likvidujeme up-dirs*/
   while(thisdir.delete())
    {
     thisdir=new File(thisdir.getParent());
     if(thisdir==null) break;
    }
    localdir=null;
    return rc;
}

final public int countObjects()
{
 return objects.size();
}

final public synchronized cacheobject getObject(String name)
{
 cacheobject obj;
 obj=(cacheobject) objects.get(name);
 if(obj!=null) return obj;
 obj=new cacheobject(name,this);
 objects.put(name,obj);
 return obj;
}

final synchronized void putObject(cacheobject co)
{
 dirty=true;
 objects.put(co.getName(),co);
}

final public boolean equals(Object o)
{
 if(o==null || ! (o instanceof cachedir)) return false;
 cachedir o1=(cachedir)o;
 
 return localdir.equals(o1.getLocalDir());
}

final public String getLocalDir()
{
 return localdir;
}

final public String toString()
{
 return localdir;
}

final public int hashCode()
{
 if(localdir==null) return 0;
 return localdir.hashCode();
} 

final public String[] listLocalNames()
{
 String res[]=new String[objects.size()];
 int i=0; 
 Enumeration en=objects.elements();
 while(en.hasMoreElements())
  {
   cacheobject o=(cacheobject)en.nextElement();
   res[i++]=o.getLocalName();
  }
  return res;
}

/* smaze neexistujici objekty */
final public void checkDir()
{
 Enumeration en=objects.elements();
 while(en.hasMoreElements())
  {
   cacheobject o=(cacheobject)en.nextElement();
   if(!o.isValid()) { objects.remove(o.getName()); 
                      dirty=true;
                      System.out.println("Missing object: "+o);
                      }
  }
}

/* smaze objekt */
final public void remove(cacheobject o)
{
 objects.remove(o.getName());
 o.delete();
 dirty=true;

}

final public Enumeration getObjects()
{
 return objects.elements();
}

} /* class */