/*
 *  Smart Cache , http proxy caching server
 *  Copyright (C) 1998-9 Radim Kolar <hsn@cybermail.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA or
 *  download copy from http://www.gnu.org/copyleft/gpl.html
 *
 */

import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.net.*;

public final class cacheobject{

/* statistics */
public static int c_hit,c_miss,c_refresh,c_block;
public static int b_hit,b_miss;

/* static ! */
public static boolean end_dot;
public static boolean keep_deleted;
public static int generate_lastmod;
public static boolean hide_errors;
public static boolean escape_backslash;
public static String custom400;
public static String defaultname;

private String name;      /* real! */
private String localname;
private String location; /* for redirects */

private int size,httprc;
private long expires, date, lastmod, lru;
private cachedir dir;
private String ctype;

cacheobject(String name, cachedir where)
{
 this.name=name;
 dir=where;
}

cacheobject(String name,cachedir where,String ln,long lm,String ct, int size)
{
 //if(name.equals(mgr.DEFAULTNAME)) name="";
 this(name,where);
 localname=ln;
 date=lastmod=lm;
 lru=System.currentTimeMillis();
 ctype=ct;
 this.size=size;
 httprc=200;
 // dir.setDirty();
}

final public boolean equals(Object o)
{
 if(o==null || ! (o instanceof cacheobject)) return false;
 cacheobject o1=(cacheobject)o;
 
 return name.equals(o1.getName());
}

final public int hashCode()
{
 return name.hashCode();
}

final public boolean needSave()
{
 if(localname!=null) return true;
  else
 return false;
}

final public String getName()
{
 return name;
}
cacheobject(DataInputStream is, cachedir d) throws IOException
{
 this.dir=d;
 name=is.readUTF();
 ctype=is.readUTF();
 localname=is.readUTF();
 location=is.readUTF();
 if(location.length()==0) location=null;
 size=is.readInt();
 httprc=is.readInt();
 expires=is.readLong();
 date=is.readLong();
 lastmod=is.readLong();
 lru=is.readLong(); 
 /* FIX DATE
 long now=System.currentTimeMillis(); 
 if(date>=now) { date=now; d.setDirty();
                 System.out.println("Fixing date (date>now) on "+d.getLocalDir()+localname);
                 // d.save();
                 }
 */                 
}

final public void save(DataOutputStream os) throws IOException
{
 if(localname==null) return;
 
 os.writeUTF(name);
 os.writeUTF(ctype);
 os.writeUTF(localname); 
 if(location!=null) os.writeUTF(location); 
  else 
    os.writeUTF("");
 os.writeInt(size);
 os.writeInt(httprc);
 os.writeLong(expires);
 os.writeLong(date); 
 os.writeLong(lastmod); 
 os.writeLong(lru);
}

final public void make_request(request r, long reload_age,long min_age, long max_age, float percent) throws IOException
{
 lru=System.currentTimeMillis();
 
 /* novy objekt */
 if(localname==null) { load_object(r); return;}
 
 /* test na expiraci */
 if(needRefresh(r,reload_age,min_age,max_age,percent)) { refresh_object(r); return;}
 
 /* jinak poslat z cache */
 send_fromcache(r);
}


final private File genTmp()
{
 File out;
 String ld=dir.getLocalDir();
 if(ld==null) return null;
 while(true)
 {
  out=new File(ld+(int)(Math.random()*100000000f)+".tmp");
  if(!out.exists()) break;
 }
 return out;
}

final public boolean isValid()
{
  if(localname==null) return false;
  File f=new File(  dir.getLocalDir()+localname );
  if(f.isDirectory()) return false;
  return f.canRead();
}

final public String getLocalName()
{
 return localname;
}

final private synchronized String genName()
{
 /* jestlize mame ? a nejaky bordel za nim, uriznout! */
 String tmp;
 tmp=name;
 File f;
 String localdir=dir.getLocalDir();
 
 if(localdir==null) return null;
 // System.err.println("in="+name);
    /* new code */
   int j=tmp.length()-1;
   if(j>=0)
   {
     byte v[];
     v=new byte[j+1];
     tmp.getBytes(0,j+1,v,0);
     loop1:for(int zz=0;zz<=j;zz++)
       {
        switch(v[zz])
         {
          case 0x3b: // ;
          case 0x3a: // :
          case 0x3d: // =
          case 0x3f: // ?
          tmp=tmp.substring(0,zz);
          // System.err.println("new tmp="+tmp);
          break loop1;
         }
       }
   }
 // System.err.println("in2="+tmp);   
 /* old - kod 
 int i=tmp.indexOf('?',0);
 if(i>-1) tmp=tmp.substring(0,i);
 //  nejaci kreteni pouzivaji i stredniky 
 i=tmp.indexOf(';',0);
 if(i>-1) tmp=tmp.substring(0,i);
 */
 
 /* pokud je toto redirect, pridat k jmenu _r */
 /* aby se to nemotalo do jmena adresare */
 if(httprc==301 || httprc==302) tmp+="_r";
 
  /* nahrazujeme nepratelske znaky */
  // tmp=tmp.replace(':','_'); neni uz nutne
  // tmp=tmp.replace('~','-');
  tmp=tmp.replace('>','}');  
  tmp=tmp.replace('<','{');
  if(escape_backslash) tmp=tmp.replace('\\','-');
  if(tmp.length()==0) tmp=defaultname;
  
  /* Kill tecku if nedovoleno ! */
  if(!end_dot && tmp.endsWith(".")) tmp=tmp.substring(0,tmp.length()-1);
  
  f=new File(localdir+tmp);  
  if(!f.exists()) return tmp;
  /* a nyni kolotoc! */
  int i=0;
  while(true)
  {
   f=new File(localdir+i+tmp);
   if(!f.exists()) break;
   i++;
   // if(i==1000) return null; // need to check??
  } 
  return i+tmp;
}


/* L O A D ! */

final private void load_object(request r) throws IOException
{
 c_miss++;
 if(localname==null) localname="<>"; /* prevents from async dir cleaning */
 if(ctype==null) ctype="unknown/unknown";
 // httprc=500;
 String ln;
 r.add_ims();
 Socket toserver=null;
 dir.setDirty();
 try{
 toserver=r.connectToHost();
 }
 catch (IOException ccc) { if(localname.equals("<>")) localname=null;
                            dir.setDirty();
                            /* POSLAT NOT MODIFIED! */
                            if(r.getIms()>0) { r.make_headers(304,null,null,0,0,0);
                                               r.send_headers();
                                               return;
                                              } 
                                              
                          /* 400 RC komedie */                    
                          if(custom400==null) r.send_error(400,"Connect to target failed ("+ccc+") and object not found in cache");
                              else
                          if(custom400.equals("-")) { 
                                                     r.send_error(204,"No Content");  
                                                    }
                          else if(custom400.equals("!"))
                          {
                           r.make_headers(200,"text/html",null,0,new Date().getTime(),0);
                           r.send_headers();
                           r.sendString("<HTML></HTML>");
                          }                                       
                        else
                        { r.make_headers(301,null,custom400,0,new Date().getTime(),0);r.send_headers();}
                
                       r.close();
                       return;
                      }
                                              
                           
 DataInputStream sin=new DataInputStream (new BufferedInputStream(toserver.getInputStream()));
 DataOutputStream sou=new DataOutputStream(new BufferedOutputStream(toserver.getOutputStream()));
 try
 {
 r.send_request(sou);
 r.read_headers(sin);
 r.send_headers();
 }
 catch (IOException e) {if(localname.equals("<>")) localname=null;dir.setDirty();toserver.close();return;}
 httprc=r.getRc();
 if(httprc==304) { localname=null;dir.setDirty();sin.close();sou.close();toserver.close();return;}
 
 File f=null;
 switch(httprc)
 {
	case 200:		/* OK */
	case 203:		/* Non-Authoritative Information */
	case 300:		/* Multiple Choices */
	case 301:               /* Moved Permanently */
        case 302:		/* Moved Temporary */
	case 410:		/* Gone */
                 f=genTmp();
 }
 if(r.requestReload()) f=null;
 DataOutputStream fout=null; 
 
 try{
 if(f!=null) fout=new DataOutputStream(
                       new BufferedOutputStream(new FileOutputStream(f),4096));
             else
               { fout=null;localname=null;dir.setDirty();}
 }
 catch (IOException e) { fout=null;localname=null;dir.setDirty();}
 
 try{
 r.transfer_object(sin,fout,dir);
 }
 catch (IOException e)
  {
   if(f!=null) f.delete();
   if(localname!=null && localname.equals("<>")) localname=null;
   dir.setDirty();
   throw e;
  }
 sin.close();sou.close();toserver.close();
 if(f==null) { localname=null;dir.setDirty();return;}
 /* nenahrali jsme jiz tento objekt? */
 if(localname==null || localname.equals("<>")) ln=genName();
   else
    ln=localname;
 File n=new File(dir.getLocalDir()+ln);
 if(!rename(f,n)) { System.err.println("Load Error: Rename "+f+" to "+n+" failed."); localname=null;dir.setDirty();return;}
 /* update flagy! */
 localname=ln;
 setInfo(r);
 System.out.println("New object: "+n+" (size="+size+" B)");
 b_miss+=size;
}

/* R E F R E S H ! */

final private void refresh_object(request r) throws IOException
{
 // System.out.println("Refreshing object: "+dir.getLocalDir()+localname);
 /* spojit se se serverem */
 Socket toserver=null;
 dir.setDirty(); 
 try{
 toserver=r.connectToHost();
 }
 catch (IOException eee) { send_fromcache(r);return;}
 
 long clims=r.getIms();
 if(clims==0) clims=1; /* nema nic, to je to same jako kdyby mel hodne starou verzi */
                       /* pokud by se nechalo na 0 - nedostal by OBJEKT! */
 if (clims>=lastmod) clims=0; /* ma pozdejsi verzi objektu nez my, posuzujeme to jako
                                 kdyby mel tu samou jako my, protoze IMS delame na nas objekt */
 r.add_ims(lastmod);
  
 DataInputStream sin=new DataInputStream (new BufferedInputStream(toserver.getInputStream()));
 DataOutputStream sou=new DataOutputStream(new BufferedOutputStream(toserver.getOutputStream()));
 
 try{
 r.send_request(sou);
 r.read_headers(sin);
 }
 catch (IOException e)
   {send_fromcache(r); return;}
   
 /* podivame se co prislo, pokud !=304, tak to posleme a nacachujeme */
 int rc=r.getRc(); 
 if(rc==304) { 
               date=System.currentTimeMillis();
               dir.setDirty();
               toserver.close();
               expires=r.getExpires(); // refresh possible Expires for Microsoft IIS 4.0
               System.out.println("Checked object: "+dir.getLocalDir()+localname);
               if(clims>0) { /* musime poslat objekt z cache */
                             r.clearIms();
                             send_fromcache(r);
                             return;}
               r.send_headers();     
               c_hit++;                         
               return;
             }
  c_refresh++;
             /* likvidujeme objekt! */ 
 if(rc==404) 
    if(keep_deleted==false) { new File(dir.getLocalDir()+localname).delete();
               System.out.println("Deleted object: "+dir.getLocalDir()+localname);
               localname=null;
               dir.setDirty();
              }
    else
     {
      System.out.println("Object "+dir.getLocalDir()+localname+" doesn't longer exist - sending cached copy");
      date=System.currentTimeMillis();
      dir.setDirty();
      try{
          toserver.close();
         }
      catch (IOException e) {}
         
      send_fromcache(r);
      return;
     }
 if( (rc==400 || rc==403 || rc>= 500) && hide_errors==true)
  {
      System.out.println("Remote server error on object "+dir.getLocalDir()+localname+" - sending cached copy");
      try{
          toserver.close();
         }
      catch (IOException e) {}
      send_fromcache(r);
      return;
  }
 r.send_headers(); 
  
 /* musime to ulozit na disk! */
 File f=null;
 switch(rc)
 {
	case 200:		/* OK */
	case 203:		/* Non-Authoritative Information */
	case 300:		/* Multiple Choices */
	case 301:               /* Moved Permanently */
        case 302:		/* Moved Temporary */
	case 410:		/* Gone */
                 f=genTmp();
 }
 
 if(r.requestReload()) f=null; /* pragma no-cache */
 DataOutputStream fout;
 
 try{
 if(f!=null) fout=new DataOutputStream(
                       new BufferedOutputStream(new FileOutputStream(f),4096));
             else
               fout=null;
 }
 catch (IOException e) {fout=null;}              
 try{
 r.transfer_object(sin,fout,dir);
 }
 catch (IOException e)
  {
   if(f!=null) f.delete();
   throw e;
  }
 sin.close();sou.close();toserver.close();
 if(f==null) return;
 /* stalled entry? - regenerate object name */
 if(localname==null || localname.equals("<>")) localname=genName();
 
 File n=new File(dir.getLocalDir()+localname);

 /* rename to new */
 if(!rename(f,n)) { 
  System.out.println("Refresh Error: Rename "+f+" to "+n+" failed.");
  dir.setDirty();return;}
 /* update hlavicek */
 clims=lastmod;
 setInfo(r);
 if(lastmod!=0) {
                  long delta=(lastmod-clims)/60000L;
                  if(delta>60*48) System.out.println("Refreshed object: "+n+" (delta="+delta/60/24+" d, size="+size+" B)");
                    else
                  if(delta>200) System.out.println("Refreshed object: "+n+" (delta="+delta/60+" h, size="+size+" B)");                    
                    else
                   System.out.println("Refreshed object: "+n+" (delta="+delta+" m, size="+size+" B)");
         }
  else System.out.println("Refreshed object: "+n+" (size="+size+" B)");
  b_miss+=size;
}

final private synchronized boolean rename(File from, File to)
{
 to.delete();
 boolean r=from.renameTo(to);
 if(!r && !to.exists()) localname=null;
 if(!r) from.delete();
 return r;
}

/*    S E N D    F R O M    C A C H E    */

final private void send_fromcache(request r) throws IOException
{
 dir.setDirty();
 if(r.getIms()>0)  { c_hit++;
                     r.make_headers(304,null,null,0,0,0);
                     r.send_headers();
                     return;
                    }
 
 /* otevrit soubor */
 DataInputStream fin=null;
 try{
 fin=new DataInputStream(
                       new BufferedInputStream(new FileInputStream(dir.getLocalDir()+localname),4096));
 }
 catch (IOException e) { System.err.println("Missing object: "+dir.getLocalDir()+localname);
                         localname=null;
                         load_object(r);
                         return;}
 c_hit++;
 b_hit+=size;
 if(generate_lastmod>0)
    r.make_headers(httprc,ctype,location,size,lastmod==0? date:lastmod,expires);
   else
    r.make_headers(httprc,ctype,location,size,lastmod,expires);
 r.send_headers();
 if(r.getMethod()==httpreq.REQUEST_GET) r.transfer_object(fin,null,dir);
}

final private void setInfo(request r)
{
 httprc=r.getRc();
 ctype=r.getCtype();
 location=r.getLocation();
 size=r.getCsize();
 lastmod=r.getIms();
 expires=r.getExpires();
 date=System.currentTimeMillis();
 /*
 if(lastmod>=date) 
   {
   System.out.println("Warning Lastmod is in future! on "+dir.getLocalDir()+localname);
   if(lastmod-date<1000*60*180) date=lastmod+1000l;
     else
      lastmod=0;
   }
 */
 dir.setDirty();
}

final private boolean needRefresh(request r,long reload_age,long min_age, long max_age, float percent)
{
 // System.out.println("Checking "+dir.getLocalDir()+localname);
 long now=System.currentTimeMillis();
 long age=now-date;
 /* - test reload status */
 if(r.requestReload() && age>reload_age) 
   {
    // System.out.println("Refreshed on user request");
    return true;
   }
 
 /* SQUID RULEZ! */
 if(age<=min_age) 
                 {
                  // System.out.println("No Refresh: AGE<= MINAGE "+age/60000+" "+min_age/60000);
                  return false;
                  
                  }
 if(age>max_age) 
       {
         // System.out.println("Refresh: AGE> MAXAGE "+age/60000+" "+max_age/60000);
         return true; /* puvodne to bylo az za expires ... */
       }
 
 
 if(expires!=0) 
   {
   if(expires<=now) 
      {
        // System.out.println("Expired...");
        return true;
      }
    else
     {
     // System.out.println("Not Yet Expired...");
     return false;
     }
   }
     
 if(lastmod==0) return true; /* usetrime si nejake to pocitani */
 long lm_age=date - lastmod;
 float lmfactor=(float) age / (float)lm_age;    
 // System.out.println("lmfactor="+lmfactor+" percent="+percent);
 
 if(lmfactor<percent && lmfactor>0) return false;
  else
 return true; /* refresh it! */
}

final public String toString()
{
 return dir.getLocalDir()+name;
}

final public cachedir getDirectory()
{
 return dir;
}

final public long getLRU()
{
 return lru;
}

final public int getSize()
{
 return size;
}

final public long getExp()
{
 return expires;
}

final public void delete()
{
 if(localname==null) return;
 System.out.println("Deleted object: "+dir.getLocalDir()+localname);
 new File(dir.getLocalDir()+localname).delete();
}

final public boolean isCheckable()
{
 if(lastmod==0) return false; else return true;
}
} /* class */