/*
 *  Smart Cache , http proxy caching server
 *  Copyright (C) 1998-9 Radim Kolar <hsn@cybermail.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA or
 *  download copy from http://www.gnu.org/copyleft/gpl.html
 *
 */

import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.net.*;

public final class garbage
{
private int cachesize;
private int dirsize;
private int oursize;
private int lowmark,highmark;
private int maxsize,refage;
private int blocksize;

private cacheobject gcarray[];
private float gcvalues[];

private int gcused;
private float gclow;

private long now;

private int lopsize,lopsizeinc;
private String lopact1,lopact2;

private int sopsize;
private String sopact;
private String exppenal, notchpenal,ebc;
private regexp urlmask[];
private String urlact[];

String root;

garbage(String root)
{
 this.root=root;
 /* init to defaults */
 cachesize=20*1024*1024; /* 20MB Cache */
 maxsize=1024*1024;
 lowmark=90;
 highmark=95;
 refage=365;
 blocksize=4096;
 dirsize=blocksize*2; 
 lopsize=1024*1024*10;
 gcused=0;
 gclow=0;
 gcarray=new cacheobject[1000];
 gcvalues=new float[1000];
 exppenal="*100";
 ebc="*4";
 /* */
 read_config("gc.cnf");
 urlmask=fixup_urls(urlmask);
 /* reverze it! */
 urlmask=mgr.reverseRegexpArray(urlmask);
 urlact=mgr.reverseStringArray(urlact);
 now=System.currentTimeMillis();
 System.out.println(new Date()+" Smart Cache garbage collector 0.8 initialized");
}

private final void read_config(String cfgfile)
{
 try{
 String line,token;
 StringTokenizer st;
 int lineno=0;
 DataInputStream dis=new DataInputStream(new BufferedInputStream(new FileInputStream(cfgfile))); 
 while ( (line = dis.readLine()) != null)
 {
          lineno++;
          if(line.startsWith("#")) continue;
	  st=new StringTokenizer(line);
          if(st.hasMoreTokens()==false) continue;
          token=st.nextToken();
          try
          {
          
          if(token.equalsIgnoreCase("cache_size")) {cachesize=Integer.valueOf(st.nextToken()).intValue()*1024*1024;
          					    continue;
                                                    }

          if(token.equalsIgnoreCase("average_directory_size")) {dirsize=Integer.valueOf(st.nextToken()).intValue();
          					    continue;
                                                    }
                                                    
          if(token.equalsIgnoreCase("cache_low_mark")) {lowmark=Integer.valueOf(st.nextToken()).intValue();
          					    continue;
                                                    }

          if(token.equalsIgnoreCase("cache_high_mark")) {highmark=Integer.valueOf(st.nextToken()).intValue();
          					    continue;
                                                    }
          if(token.equalsIgnoreCase("block_size")) {blocksize=Integer.valueOf(st.nextToken()).intValue();
          					    continue;
                                                    }
                                                    
          if(token.equalsIgnoreCase("gcarraysize")) {gcarray=new cacheobject[Integer.valueOf(st.nextToken()).intValue()];
                                                     gcvalues=new float[gcarray.length];
          					    continue;
                                                    }
                                                    
          if(token.equalsIgnoreCase("maximum_object_size")) {
                                                    maxsize=Integer.valueOf(st.nextToken()).intValue()*1024;
          					    continue;
                                                    }
          if(token.equalsIgnoreCase("reference_age")) {
                                                    refage=Integer.valueOf(st.nextToken()).intValue();
          					    continue;
                                                    }

          if(token.equalsIgnoreCase("large_object_penalty")) {
                                                    lopsize=Integer.valueOf(st.nextToken()).intValue();
                                                    lopact1=st.nextToken();
                                                    lopsizeinc=Integer.valueOf(st.nextToken()).intValue();
                                                    lopact2=st.nextToken();
          					    continue;
                                                    }
          if(token.equalsIgnoreCase("small_object_penalty")) {
                                                    sopsize=Integer.valueOf(st.nextToken()).intValue();
                                                    sopact=st.nextToken();
          					    continue;
                                                    }
                                                    
          if(token.equalsIgnoreCase("expired_penalty")) {
                                                    exppenal=st.nextToken();
          					    continue;
                                                    }
                                                    
          if(token.equalsIgnoreCase("expired_but_checkable_penalty")) {
                                                    ebc=st.nextToken();
          					    continue;
                                                    }

          if(token.equalsIgnoreCase("not_checkable_penalty")) {
                                                    notchpenal=st.nextToken();
          					    continue;
                                                    }
                                                    
          if(token.equals("case_sensitive_matching")) {
          				                   char c=(char)Integer.valueOf(st.nextToken()).intValue();
                                                           if(c==1) mgr.case_sensitive=true; else mgr.case_sensitive=false;
                                                           continue;
                                                      }
                                                    
          if(token.equalsIgnoreCase("urlmask")) {
          					 token=st.nextToken();
                                                 if(mgr.isInRegexpArray(token,urlmask)) continue;
                                                 urlmask=mgr.addRegexpToArray(token,urlmask);
                                                 urlact=mgr.addStringToArray(st.nextToken(),urlact);
          					 continue;
                                                 }
          }
          catch (NoSuchElementException z)
           {
          System.out.println("[GC_CONFIG_ERROR] Missing argument(s) in "+cfgfile+":"+lineno);
          continue;
           }
           
          catch (NumberFormatException z)
           {
          System.out.println("[GC_CONFIG_ERROR] Invalid number in "+cfgfile+":"+lineno);
          continue;
           }

          System.out.println("[GC_CONFIG_ERROR] Unknown keyword "+token+" in "+cfgfile+":"+lineno);
 }/* while */
 dis.close();
   
 }
 catch(IOException e) {}

}

private final regexp[] fixup_urls(regexp what[])
{
 if(what==null) return null;
 for(int i=0;i<what.length;i++)
  {
  regexp r;
  r=new regexp( encodechars(what[i].toString()),false);
  what[i]=r;
  }
  return what;
}

public final void cleanup()
{
 kill_unref_dir(root,true);
}

public final void converto020()
{
 convert020dir(root);
}

public final void converto030()
{
 convert030dir(root);
}

private final void dump_gcarray()
{
 System.out.println("\n"+new Date()+" Dumping garbage collection memory array");
 for(int i=gcused;i>=0;i--)
  {
   if(gcarray[i]==null) continue;
   System.out.println(gcvalues[i]+"  "+gcarray[i].getDirectory().getLocalDir()+gcarray[i].getName());
  }
}

private final void convert020dir(String dirname)
{
  File root=new File(dirname);
  
  String dirfilez[];
  Vector filez=new Vector();
  
  if(!root.isDirectory()) return;
  
//  if(!new File(dirname,cachedir.DIRINFO).exists()) return;
  try{
  
  DataInputStream dis=new DataInputStream( new BufferedInputStream(new FileInputStream(root+File.separator+cachedir.DIRINFO)));
  
  DataOutputStream dos=new DataOutputStream( new BufferedOutputStream(new FileOutputStream(root+File.separator+cachedir.DIRINFO+"new")));
  
  int howmany=dis.readInt();
  dos.writeInt(howmany);
  
  for(int i=0;i<howmany;i++)
  {
  String s;
  /* name */
  s=dis.readUTF();
  dos.writeUTF(s);
  /* ctype */  
  s=dis.readUTF();
  dos.writeUTF(s);
  /* localname */  
  s=dis.readUTF();
  dos.writeUTF(s);
  /* location */
  s=dis.readUTF();
  dos.writeUTF(s);
  int x;
  /* size */
  x=dis.readInt();
  dos.writeInt(x);
  /* httprc */
  x=dis.readInt();
  dos.writeInt(x);
  long l;
  /* expires */
  l=dis.readLong();
  dos.writeLong(l);
  /* date */
  l=dis.readLong();
  dos.writeLong(l);
  /* lastmod */
  l=dis.readLong();
  dos.writeLong(l);
  /* LRU */
  dos.writeLong(System.currentTimeMillis());
      
  }
 dis.close(); 
 dos.close();  
 new File(dirname,cachedir.DIRINFO).delete();
 if(!new File(dirname,cachedir.DIRINFO+"new").renameTo(new File(dirname,cachedir.DIRINFO)))
   System.out.println("Rename failed in dir "+dirname);
 }
  catch (FileNotFoundException e1) {}
  catch (Exception e) {System.out.println(e);}
  
  dirfilez=root.list();
  /* rekurzivne do adresaru */
  for(int i=0;i<dirfilez.length;i++)
   {
     if(!new File(root,dirfilez[i]).isDirectory()) continue;
     convert020dir(dirname+File.separator+(String)dirfilez[i]);
   }  
}



private final void convert030dir(String dirname)
{
  File root=new File(dirname);
  
  String dirfilez[];
  Vector filez=new Vector();
  boolean goodlevel=false;
  if(!root.isDirectory()) return;
  
  /* zjistit,  zda jsme na spravnem levelu t.j. d:\store\x\y\zzzzz.yy */
  if(dirname.length()-this.root.length()>5)
  {
  String nd=dirname.substring(this.root.length()+1);
  int i=nd.indexOf(File.separatorChar,0);
  if(i!=-1) {
  nd=nd.substring(i+1);
  i=i=nd.indexOf(File.separatorChar,0);
  if(i!=-1) goodlevel=true;
  }
  // System.out.println(dirname+" = "+nd);
  
  }
  
  /* rekurzivne do adresaru */
  dirfilez=root.list();
  for(int i=0;i<dirfilez.length;i++)
   {
     if(!new File(root,dirfilez[i]).isDirectory()) continue;
     if(goodlevel==true && dirfilez[i].startsWith("-"))
     {
      new File(root,dirfilez[i]).renameTo(new 
       File(root,"~"+dirfilez[i].substring(1))  );
     }
     if(goodlevel==false) convert030dir(dirname+File.separator+(String)dirfilez[i]);
   }  
}


/* ************************************* */
/* ALERT: Synchronize by hand with mgr!! */
/* ************************************* */

/* TODO: support for other protocols...  */

private final String encodechars(String urldir)
{
 /* odstranit http:// if any */
  // System.out.println("Encode in="+urldir);
  /* get protocol if any */
  int i=urldir.indexOf("://",0);
  if(i!=-1)
   {
    String proto=urldir.substring(0,i);
    urldir=urldir.substring(i+3); //cut of protocol
    if(!proto.equalsIgnoreCase("http"))
     {
      /* ftp://ftp.debian.org/* -> ftp.debian.org^ftp/* */
      /* ftp://ftp.lamer* -> ftp.lamer*^ftp             */
      
      i=urldir.indexOf("/",0);
      if(i==-1)  urldir+='^'+proto+'*';
        else
      urldir=urldir.substring(0,i)+'^'+proto+urldir.substring(i);
     }
   }
  // if(urldir.startsWith("http://")) urldir=urldir.substring(7);

  /* nahrazujeme nepratelske znaky */
  urldir=urldir.replace(':','_');
  //  urldir=urldir.replace('*','@');
  //  urldir=urldir.replace('?','#');  
  // urldir=urldir.replace('~','-');
  // urldir=urldir.replace('>','}');  
  // urldir=urldir.replace('<','{');
  urldir=urldir.replace('|','!');
  if(File.separatorChar!='/') urldir=urldir.replace('/',File.separatorChar);
  // System.out.println("encode debug="+urldir);
  return urldir;
}

public final void garbage_collection()
{
 int bot=(int)( (float)cachesize*(float)lowmark/100f);
 int high=(int)( (float)cachesize*(float)highmark/100f);
 System.out.println(new Date()+" Cache size watermarks: Low="+bot+" ("+lowmark+"%) High="+high+" ("+highmark+"%)");
 // boolean cleaning=false;
 oursize=high+1;
 while(oursize>high)
 {
 oursize=0;
 gcarray=new cacheobject[gcarray.length];
 gcvalues=new float[gcvalues.length];
 System.out.println(new Date()+" Scanning cache directory. This may take some time...");
 scan_cache(root);
 System.out.println(new Date()+" Scan end. Cachesize="+oursize+
 " B\n"+new Date()+" GC array time marks: Low ="+gclow+" high="+
 (gcused>=gcvalues.length ? new Float(gcvalues[gcvalues.length])
 : new Float(gcvalues[gcused])));
 // dump_gcarray();
 delete_files();
 // if(cleaning==false && oursize<high) break; else cleaning=true; 
 }
} 

public final void fake_garbage_collection()
{
 refage=Integer.MAX_VALUE;
 int bot=(int)( (float)cachesize*(float)lowmark/100f);
 int high=(int)( (float)cachesize*(float)highmark/100f);
 System.out.println(new Date()+" Cache size watermarks: Low="+bot+" ("+lowmark+"%) High="+high+" ("+highmark+"%)"); 
 // boolean cleaning=false;
 oursize=0;
 gcarray=new cacheobject[gcarray.length];
 gcvalues=new float[gcvalues.length];
 System.out.println(new Date()+" Scanning cache directory. This may take some time...");
 scan_cache(root);
 System.out.println(new Date()+" Scan end. Cachesize="+oursize+
 " B\n"+new Date()+" GC array time marks: Low ="+gclow+" high="+
 (gcused>=gcvalues.length ? new Float(gcvalues[gcvalues.length])
 : new Float(gcvalues[gcused])));
 dump_gcarray();
 fake_delete_files();
 // if(cleaning==false && oursize<high) break; else cleaning=true; 
} 

private final void delete_files()
{
 int bot=(int)( (float)cachesize*(float)lowmark/100f);
 int high=(int)( (float)cachesize*(float)highmark/100f);
 if(oursize<bot)
  {
   System.out.println(new Date()+" Cache size is OK - No further cleaning needed");
   return;
  }
 
 System.out.println(new Date()+" Cleaning files....");
 for(int i=gcused;i>=0;i--)
 {
  if(gcarray[i]==null) continue;
  /* smazat soubor */
  gcarray[i].getDirectory().remove(gcarray[i]);
  int size;
  size=gcarray[i].getSize();
  oursize-= (size/blocksize)*blocksize;
  if(size % blocksize !=0 ) oursize-=blocksize;
  
  if(gcarray[i].getDirectory().countObjects()==0)
   {
    /* TODO half dir! */
    oursize-=dirsize;
    gcarray[i].getDirectory().cleandir();
   }
   if(oursize<bot) break;
 }
 /* SAVE Directories */
 System.out.println(new Date()+" Updating directory information...");
 for(int i=gcused;i>=0;i--)
 {
  if(gcarray[i]==null) continue;
  gcarray[i].getDirectory().save();
  
 } 
 System.out.println(new Date()+" Done. Cachesize is now "+oursize+" bytes.");
 if(oursize>high) System.out.println(new Date()+" WARNING: Another rescan cache needed to finish GC. If this happens frequently increase gcarraysize in gc.cnf");
}

private final void fake_delete_files()
{
 int bot=(int)( (float)cachesize*(float)lowmark/100f);
 int high=(int)( (float)cachesize*(float)highmark/100f);
 if(oursize<bot)
  {
   System.out.println(new Date()+" [FAKE] Cache size is OK - No further cleaning needed");
   return;
  }
 
 System.out.println(new Date()+" [FAKE] Not Cleaning files....");
 for(int i=gcused;i>=0;i--)
 {
  if(gcarray[i]==null) continue;
  /* smazat soubor */
  System.out.println("[FAKE] Delete "+gcarray[i].getDirectory()+gcarray[i].getName());
  int size;
  size=gcarray[i].getSize();
  oursize-= (size/blocksize)*blocksize;
  if(size % blocksize !=0 ) oursize-=blocksize;
  
  if(oursize<bot) break;
 }
 
 System.out.println(new Date()+" [FAKE] Done. Cachesize is now "+oursize+" bytes.");
 if(oursize>high) System.out.println(new Date()+" WARNING: Another rescan cache needed to finish GC.");
}

/* proscanuje rekurzivne cache a naplni gcarray */
private final void scan_cache(String dirname)
{
 oursize+=dirsize;
 cachedir cd;
 /* nacist adresar */
 cd=kill_unref_dir(dirname,false);
 /* nacpat do gc pole */
 if(cd!=null)
  {
   /* prevest localname na URL */
   /* ze d:\store\XX\YY\name */
   /* udelat name */
   String urlpart;
   urlpart=cd.getLocalDir().substring(root.length()+1);
   /* preskocit 2x file-separator */
   int i;
   i=urlpart.indexOf(File.separatorChar,0);
   if(i!=-1) {
      urlpart=urlpart.substring(i+1); 
      i=urlpart.indexOf(File.separatorChar,0);
      if(i!=-1) urlpart=urlpart.substring(i+1); 
                else 
             { System.out.println(new Date()+" Scanning "+dirname); urlpart="";}
      } else { urlpart="";}
   
   // System.out.println(cd.getLocalDir()+"="+urlpart);
   
   Enumeration en;
   en=cd.getObjects();
   while(en.hasMoreElements())
   {
   cacheobject obj;
   obj=(cacheobject)en.nextElement();
   /* spocitat vahu objektu */
   float w;
   w=(now-obj.getLRU())/86400000.0000f;
   /* large obj penalty */
   int size;
   size=obj.getSize();
   if(size>maxsize)
                    {  cd.remove(obj);continue; }
   oursize+= (size/blocksize)*blocksize;
   if(size % blocksize !=0 ) oursize+=blocksize;
   
   if(size>lopsize) { size-=lopsize;w=applyMod(w,lopact1);
                      while(size>lopsizeinc)
                       {
                        size-=lopsizeinc;
                        w=applyMod(w,lopact2);
                       }
                    } /* LOP HACK */
  else
   if(size<=sopsize) w=applyMod(w,sopact);
   
  long xsize;
  xsize=obj.getExp();
  boolean chk;
  chk=obj.isCheckable();
  if(xsize>0 && xsize<now) 
   if(chk) w=applyMod(w,ebc); else
           w=applyMod(w,exppenal);
   else
    if (chk==false) w=applyMod(w,notchpenal);
  
  // System.out.println(dirname+obj.getName()+" w="+w);
  
  /* URL masky */
  if(urlmask!=null)
  {
  for(size=urlmask.length-1;size>=0;size--)
   if(urlmask[size].matches(urlpart+obj.getName())) { w=applyMod(w,urlact[size]);break;}
  }
   
  if(w>=refage) {
                size=obj.getSize();
                oursize-= (size/blocksize)*blocksize;
                if(size % blocksize !=0 ) oursize-=blocksize;
                
                cd.remove(obj);continue; }
  
  /* pridat do gc pole */
  if(w<gclow) continue;
  // if(gcused==-1) { gcused=0;gcarray[0]=obj;gcvalues[0]=w;continue;}
  /* najit misto pro vlozeni pomoci puleni intervalu */
  int l,h;
  l=0;
  h=gcused;
  while (l!=h)
  {
    int z;
    z=(l+h)/2;
    if(w<gcvalues[z]) {h=z;continue;}
    if(w>gcvalues[z]) {l=z+1;continue;}
    l=z;break;
  }
  /* pridat za pozici l */
  if(w<gcvalues[l]) l--;
  
  /* pridat doprava */
     if(gcused<gcvalues.length-1)
      {
       /* jeste mame misto - posunuje se doprava*/
       // System.out.println("right l="+l+" gcused="+gcused);
       System.arraycopy(gcvalues,l+1,gcvalues,l+2,gcvalues.length-l-2);
       System.arraycopy(gcarray,l+1,gcarray,l+2,gcarray.length-l-2);
       gcvalues[l+1]=w;
       gcarray[l+1]=obj;
       gcused++;
      }
     else
      {
       /* misto neni - jedeme do leva */
       // System.out.println("left l="+l+" gcused="+gcused);
       if(l>0)
       {
        System.arraycopy(gcvalues,1,gcvalues,0,l);
        System.arraycopy(gcarray,1,gcarray,0,l);
       }
       gcvalues[l]=w;
       gcarray[l]=obj;
       gclow=gcvalues[0];
      }
    
    
  } /* elements */
  /* TODO - half empty directory! */
  if(cd.countObjects()==0) { cd.cleandir();oursize-=dirsize;} else
  cd.save();
  } /* cd not null */
  
  /* dive into subdirz */
  /* potrebuje mnoho pameti!!! */
  File root=new File(dirname);
  
  String filez[];
  filez=root.list();
  if(filez==null) { oursize-=dirsize;return;}
  for(int i=filez.length-1;i>=0;i--)
   if(new File(dirname+File.separator+filez[i]).isDirectory()) scan_cache(dirname+File.separator+filez[i]);
}

private final float applyMod(float f,String mod)
{
 if(mod==null) return f;
 float m;
 try
 {
  m=Float.valueOf(mod.substring(1)).floatValue();
 }
 catch ( NumberFormatException z)
  {
   System.out.println("[GC_CONFIG_ERROR] Invalid number in LRU modifier '"+mod+"'");
   throw z;
  }
 switch(mod.charAt(0))
 {
  case '+':f+=m;break;
  case '-':f-=m;break;
  case '*':f*=m;break;
  case '/':f/=m;break;
  case '=':if(f>=m) f=refage;
           break;
  default:
   System.out.println("[GC_CONFIG_ERROR] Invalid action in LRU modifier '"+mod+"'");  
   throw new IllegalArgumentException("Unknown action "+mod.charAt(0));
 }
 return f;
}

private final cachedir kill_unref_dir(String dirname, boolean recurse)
{
  File root=new File(dirname);
  
  String dirfilez[];
  String local[];
  Vector filez=new Vector();
  
  if(!root.isDirectory()) return null;
  dirfilez=root.list();
  cachedir cd=new cachedir(dirname+File.separator);
  cd.checkDir();
  local=cd.listLocalNames();
  cd.save();
  /* konverze dirfilez na vector */
  for(int i=dirfilez.length-1;i>=0;i--)
   filez.addElement(dirfilez[i]);
  dirfilez=null;

  /* smazat localnames z filez, abychom je nesmazali */
  for(int i=local.length-1;i>=0;i--)
   filez.removeElement(local[i]);
   
  /* abychom si nesmazali cacheinfo !! */
  /* pokud jsme empty dir, tak ho smazeme */
     
  if(cd.countObjects()>0 ) filez.removeElement(cachedir.DIRINFO);
  if(recurse==true) cd=null;
    
  local=null;
  /* projedeme filez a smazeme nalezene soubory */
  for(int i=filez.size()-1;i>=0;i--)
   {
    File df=null;
    try{
    df=new File(dirname+File.separator+(String)filez.elementAt(i));
    }
    catch (ArrayIndexOutOfBoundsException z) {continue;}
    if(!df.isDirectory())
      {
       System.out.println("Deleting orphan file: "+df);
       df.delete();
       filez.removeElementAt(i);       
       i++;continue;
      }
   }
  if(!recurse) { root.delete();return cd;}  
  filez.trimToSize();
  
  /* rekurzivne do adresaru */
  for(int i=filez.size()-1;i>=0;i--)
   {
     kill_unref_dir(dirname+File.separator+(String)filez.elementAt(i),true);
   }  
  root.delete(); 
  return cd;
}

}